/*
 * Startup.S : processes after CPU RESET, from WinARM example
 *   by uratan! 2008.9.8
 */

/*
 * specify RAM address range, I will fill them at startup
 */
#define FILL_RAM_TOP  0x40000000
#define FILL_RAM_END  0x40008000

/*
 * Standard definitions of Mode bits and Interrupt (I & F) flags in PSRs
 */
	.set Mode_USR, 0x10
	.set Mode_FIQ, 0x11
	.set Mode_IRQ, 0x12
	.set Mode_SVC, 0x13
	.set Mode_ABT, 0x17
	.set Mode_UND, 0x1B
	.set Mode_SYS, 0x1F

	.set I_Bit, 0x80    /* when I bit is set, IRQ is disabled */
	.set F_Bit, 0x40    /* when F bit is set, FIQ is disabled */

/*
 * Pre-defined interrupt handlers that may be directly
 * overwritten by C interrupt functions
 */
	.extern	FIQ_handler
	.extern	UNDEF_handler
	.extern	PABT_handler
	.extern	DABT_handler


/*===================================================================*/

/*
 * Stack Configuration (Stack Sizes in Bytes)
 */
	.set UND_Stack_Size, 0x00000080		/* Stack for Undefined Mode */
	.set ABT_Stack_Size, 0x00000080		/* Stack for Abort Mode */
	.set FIQ_Stack_Size, 0x00000100		/* Stack for Fast Interrupt Mode */
	.set IRQ_Stack_Size, 0x00000200		/* Stack for Interrupt Mode (IRQ) */
	.set SVC_Stack_Size, 0x00000100		/* Stack for Supervisor Mode (SWI) */
	.set USR_Stack_Size, 0x00000300		/* Stack for User/System Mode */

.arm
.section .stack, "w"
.align 4
	.space	(USR_Stack_Size+3)&~3
	.space	(SVC_Stack_Size+3)&~3
	.space	(IRQ_Stack_Size+3)&~3
	.space	(FIQ_Stack_Size+3)&~3
	.space	(ABT_Stack_Size+3)&~3
	.space	(UND_Stack_Size+3)&~3
Top_Stack:


/*===================================================================*/

.text

/*
 * Exception Vectors, Mapped to Address 0.
 * Absolute addressing mode must be used.
 */
.arm
.section .vectorg, "ax"
__Vectors:
	LDR		PC, Reset_Addr
	LDR		PC, Undef_Addr
	LDR		PC, SWI_Addr
	LDR		PC, PAbt_Addr
	LDR		PC, DAbt_Addr
	.word	0					/* Reserved Vector (holds ISP checksum) */
	LDR		PC, [PC, #-0x0FF0]		/* Vector from VicVectAddr register */
	LDR		PC, FIQ_Addr

Reset_Addr:	.word	Reset_Handler		/* in myself */
Undef_Addr:	.word	UNDEF_handler		/* in main.c */
SWI_Addr:	.word	SoftwareInterrupt	/* in swi_handler.S */
PAbt_Addr:	.word	PABT_handler		/* in main.c */
DAbt_Addr:	.word	DABT_handler		/* in main.c */
IRQ_Addr:	.word	UNDEF_handler		/* (use VIC auto-vectoring) */
FIQ_Addr:	.word	FIQ_handler			/* in main.c */
			.word	0		/* rounds the vectors and ISR addresses to 64 bytes total */


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
 * system startup processes
 */

.arm
.section .init, "ax"
.global __startup
.func __startup
__startup:

Reset_Handler:
	/*
	 * fill all memory for debug
	 */
	LDR		R0, =0x77777777
	LDR		R1, =FILL_RAM_TOP
	LDR		R2, =FILL_RAM_END
lp01:
	CMP		R1, R2
	STRLO	R0, [R1], #4
	BLO		lp01

	/*
	 * Setup Stack for each mode with marking the stack area
	 */
	LDR		R0, =Top_Stack

	MSR		CPSR_c, #Mode_UND|I_Bit|F_Bit
	MOV		SP, R0

	MOV		R3, R0
	SUB		R0, R0, #UND_Stack_Size
	MOV		R2, R0
	LDR		R1, =0xDDDDDDDD		/* un'D'efined */
lp11:
	CMP		R2, R3
	STRLO	R1, [R2], #4
	BLO		lp11

	MSR		CPSR_c, #Mode_ABT|I_Bit|F_Bit
	MOV		SP, R0

	MOV		R3, R0
	SUB		R0, R0, #ABT_Stack_Size
	MOV		R2, R0
	LDR		R1, =0xAAAAAAAA		/*'A'bort */
lp12:
	CMP		R2, R3
	STRLO	R1, [R2], #4
	BLO		lp12

	MSR		CPSR_c, #Mode_FIQ|I_Bit|F_Bit
	MOV		SP, R0

	MOV		R3, R0
	SUB		R0, R0, #FIQ_Stack_Size
	MOV		R2, R0
	LDR		R1, =0xFFFFFFFF		/* 'F'IQ */
lp13:
	CMP		R2, R3
	STRLO	R1, [R2], #4
	BLO		lp13

	MSR		CPSR_c, #Mode_IRQ|I_Bit|F_Bit
	MOV		SP, R0

	MOV		R3, R0
	SUB		R0, R0, #IRQ_Stack_Size
	MOV		R2, R0
	LDR		R1, =0x11111111		/* '1'RQ */
lp14:
	CMP		R2, R3
	STRLO	R1, [R2], #4
	BLO		lp14

	MSR		CPSR_c, #Mode_SVC|I_Bit|F_Bit
	MOV		SP, R0

	MOV		R3, R0
	SUB		R0, R0, #SVC_Stack_Size
	MOV		R2, R0
	LDR		R1, =0x55555555		/* '5'VC */
lp15:
	CMP		R2, R3
	STRLO	R1, [R2], #4
	BLO		lp15

//	MSR		CPSR_c, #Mode_USR				/* Interrupts enabled */
//	MSR		CPSR_c, #Mode_USR|I_Bit|F_Bit	/* Interrupts disabled */
//	MSR		CPSR_c, #Mode_SYS				/* Interrupts enabled */
	MSR		CPSR_c, #Mode_SYS|I_Bit|F_Bit	/* Interrupts disabled */
	MOV		SP, R0

	MOV		R3, R0
	SUB		R0, R0, #USR_Stack_Size
	MOV		R2, R0
	LDR		R1, =0x33333333		/* u'3'er */
lp16:
	CMP		R2, R3
	STRLO	R1, [R2], #4
	BLO		lp16

	/*
	 * Relocate .data section (Copy from ROM to RAM)
	 */
	LDR		R1, =_text_end
	LDR		R2, =_data_start
	LDR		R3, =_data_end
lp21:
	CMP		R2, R3
	LDRLO	R0, [R1], #4
	STRLO	R0, [R2], #4
	BLO		lp21

	/*
	 * Clear .bss section (Zero init)
	 */
	MOV		R0, #0
	LDR		R1, =_bss_start
	LDR		R2, =_bss_end
lp31:
	CMP		R1, R2
	STRLO	R0, [R1], #4
	BLO		lp31

	/*
	 * Enter the C code
	 */
	B		main
.endfunc

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
 * firmware update processes
 */

	.set EXEC_BUFF_SIZE, 0x0400

.arm
.text
.section .text, "ax"
.global copy_n_exec_on_ram
.func   copy_n_exec_on_ram
copy_n_exec_on_ram:
	/*
	 * copy codes for update to RAM
	 */
	LDR		R1, =_fup_start
	LDR		R2, =_fup_end
	LDR		R3, =exec_buffer
	SUB		R0, R0, R1
	ADD		R0, R0, R3
lp51:
	CMP		R1, R2
	LDRLO	R4, [R1], #4
	STRLO	R4, [R3], #4
	BLO		lp51

	/*
	 * execute it
	 */
	MOV		PC, R0

.endfunc

.section .exec, "w"
.align 4
exec_buffer:
	.space	EXEC_BUFF_SIZE


.end
